import os
import time
from scipy.signal import detrend, butter, sosfiltfilt,decimate
import numpy as np
from nptdms import TdmsFile
from scipy.io import savemat

# Configuration 
INPUT_DIR_PATH="F:\data_Tromso2021\SML5\\"
OUTPUT_DIR_PATH="F:\data_Tromso2021_processed\SML5\\"
CHANNELS_TO_KEEP={
    "gps":[],
    "acceleration_data":
        [
        "timestamp",
        "A1.X",
        "A1.Y",
        "A1.Z",

        "A2.X",
        "A2.Y", 
        "A2.Z",

        "A3.X",
        "A3.Y", 
        "A3.Z",

        "A4.X",
        "A4.Y",
        "A4.Z",
        ],
    "metadata.sensor.mod7.ch00.name":
        [
        "timestamp",
        "direction",
        "magnitude",
        "vertical_velocity",
        "speed_of_sound"
        ],
    "metadata.sensor.mod7.ch01.name":[],
    "metadata.sensor.mod7.ch02.name":[],
    "metadata.sensor.mod7.ch03.name":[]
}
RESTTIME            = 60*3                                  # s
SAMPLING_FREQ       = 400                                   # Hz
BUTTER_ORDER        = 5                                     #
#

def main():
    count = 0
    processedcount= 0
    filecount=sum(1 for _ in os.scandir(INPUT_DIR_PATH))-1
    with os.scandir(INPUT_DIR_PATH) as it:
        timestart_global=time.time()
        searchpath=(OUTPUT_DIR_PATH+'\\40hz')
        
        for entry in it:
            count+=1
            if entry.name.endswith(".tdms") and entry.is_file() and entry.name.replace("tdms","mat") not in os.listdir(searchpath):
                processedcount+=1
                print(f"Processing file {entry.name} ({count}/{filecount})")
                
                #Actual processing step
                preprocess_file(entry.name.replace(".tdms",""),entry.path)
                
                #Track progress
                elapsed=(time.time()-timestart_global)//60
                print(f"\tProcessing success! Elapsed: {elapsed:.0f} min. ETA = {(elapsed/processedcount*(filecount-(sum(1 for _ in os.listdir(searchpath))-1))):.0f} min")
                
                #Giving the storage drive x minutes to cool down, preventing excessive heat buildup for sustained operation
                time.sleep(RESTTIME)

def preprocess_file(filename:str,file_path:str):
    """
    Parameters
    ----------
    filename: str
        namecore of the output file
    file_path: str
        path to the tdms file containing the data

    """
    #Create dataframes
    data40={}
    data20={}
    tdms_file = TdmsFile.read(file_path)
    print("\tFile read successfully")
    for group in tdms_file.groups():
        group_name = group.name
        for channel in group.channels():
            channel_name = channel.name
            # Check if channel data should be stored:
            if channel_name in CHANNELS_TO_KEEP[group_name]:
                channeldata = channel[:]
                #Give valid channel names for matlab (doesnt accept .)
                title=group_name+channel_name
                title=title.replace('.','_')
                if group_name=="acceleration_data":
                    #Preprocess acceleration signal
                    channeldata40, channeldata20=preprocess_acceleration_data(channeldata)
                #Store channel data into dataframe
                data40[title]=channeldata40
                data20[title]=channeldata20
    print("\tFile processed successfully")
    fullpath40=OUTPUT_DIR_PATH+"\\40hz\\"+filename+".mat"
    fullpath20=OUTPUT_DIR_PATH+"\\20hz\\"+filename+".mat"
    savemat(fullpath40, data40)
    savemat(fullpath20, data20)
    print(f"\tFile saved successfully at {fullpath40} and {fullpath20}")


def preprocess_acceleration_data(data:np.ndarray):
    """
    Preprocessing routine for acceleration data

    Parameters
    ----------
    data : array_like
        The signal to be downsampled, as an 1xN-dimensional numpy array.

    Returns
    -------
    processed_data40 : ndarray
        The preprocessed acceleration signal at 40Hz.
    processed_data20 : ndarray
        The preprocessed acceleration signal at 20Hz.
    """
    ##################40hz
    #Low pass
    soslow40=butter(
        BUTTER_ORDER,
        0.4*40,
        btype='low',
        output='sos',
        fs=SAMPLING_FREQ
        )
    processed_data40 = sosfiltfilt(soslow40, data)
    #Decimate
    processed_data40 = decimate(processed_data40,10)
    #Detrend
    processed_data40 = detrend(processed_data40)
    #Zero-centre
    processed_data40 -= np.mean(processed_data40)
    ##################20hz
    #Low pass
    soslow20=butter(
        BUTTER_ORDER,
        0.4*20,
        btype='low',
        output='sos',
        fs=SAMPLING_FREQ
        )
    processed_data20 = sosfiltfilt(soslow20, data)
    #Decimate (not recommended to decimate more than 10x per step)
    processed_data20 = decimate(processed_data20,2)
    processed_data20 = decimate(processed_data20,10)
    #Detrend
    processed_data20 = detrend(processed_data20)
    #Zero-centre
    processed_data20 -= np.mean(processed_data20)
    
    return processed_data40,processed_data20


main()